import os
import shutil
import sys

TESTING_DIR: str = os.path.abspath(os.path.join(os.curdir, "testing"))
PROTON_DIR_1: str = os.path.join(TESTING_DIR, "proton_dir_1")
PROTON_DIR_2: str = os.path.join(TESTING_DIR, "proton_dir_2")
USER_DIR: str = os.path.join(TESTING_DIR, "user_dir")

PROTON_BIG: str = "UMU_Proton_10"
PROTON_SMALL: str = "UMU_Proton_1"

sys.path.insert(1, os.path.join(os.path.abspath(os.curdir), "src"))


def teardown():
    shutil.rmtree(TESTING_DIR)


def setup():
    if os.path.exists(TESTING_DIR):
        teardown()

    os.mkdir(TESTING_DIR)
    os.mkdir(PROTON_DIR_1)
    os.mkdir(os.path.join(PROTON_DIR_1, PROTON_BIG))
    os.mkdir(os.path.join(PROTON_DIR_1, PROTON_SMALL))
    os.mkdir(PROTON_DIR_2)
    os.mkdir(USER_DIR)



from umu_commander import VERSION
from umu_commander.classes import Element, Group, Value


def string_to_value(value: str) -> Value:
    return Value(value)


def strings_to_values(values: list[str]) -> list[Value]:
    return [string_to_value(value) for value in values]


def _selection_set_valid(
    selection_elements: list[Element] | None, selection_groups: list[Group] | None
):
    if (selection_elements is None or len(selection_elements) == 0) and (
        len(selection_groups) == 0
        or all([len(group.elements) == 0 for group in selection_groups])
    ):
        return False
    else:
        return True


def _print_selection_group(
    elements: list[Element], enum_start_idx: int, tab: bool = True
):
    prefix: str = "\t" if tab else ""
    print(
        *[
            f"{prefix}{idx}) {element.value} {element.info}"
            for idx, element in enumerate(elements, start=enum_start_idx)
        ],
        sep="\n",
    )
    print("")


def _translate_index_to_selection(
    selection_index: int,
    selection_elements: list[Element],
    selection_groups: list[Group],
) -> Element:
    len_counter: int = 0

    if selection_elements is not None:
        selection_groups.insert(0, Group(elements=selection_elements))

    for group in selection_groups:
        len_counter += len(group.elements)
        if len_counter > selection_index:
            break

    return Element(
        group.identity, group.elements[selection_index - len_counter].value, ""
    )


def get_selection(
    prompt: str,
    selection_elements: list[Element] | None,
    selection_groups: list[Group] | None,
    default_element: Element = None,
) -> Element:
    if not _selection_set_valid(selection_elements, selection_groups):
        print("Nothing to select from.")
        raise ValueError

    if selection_groups is None:
        selection_groups = []

    while True:
        enum_start_idx: int = 1

        print(prompt)

        if selection_elements is not None:
            _print_selection_group(selection_elements, enum_start_idx, tab=False)
            enum_start_idx += len(selection_elements)

        for group_idx, group in enumerate(selection_groups):
            if len(group.elements) == 0:
                continue

            print(group.label)
            _print_selection_group(group.elements, enum_start_idx)

            enum_start_idx += len(group.elements)

        selection: str = input("? ")
        print("")
        if selection == "":
            if default_element is not None:
                return default_element

            # If only 1 choice
            if len(selection_groups) == 0 and len(selection_elements) == 1:
                return selection_elements[0]

            # len(selection_groups) == 1 and len(selection_groups[0].elements) == 1
            groups_with_one_element: list[Group] = [
                group for group in selection_groups if len(group.elements) == 1
            ]
            if len(groups_with_one_element) == 1:
                return groups_with_one_element[0].elements[0]

        elif selection.isdecimal():
            selection: int = int(selection) - 1
            if enum_start_idx - 1 > selection >= 0:
                break

    return _translate_index_to_selection(
        selection, selection_elements, selection_groups
    )


def print_help():
    print(
        f"umu-commander {VERSION}",
        "Interactive CLI tool to augment umu-launcher as well as help you manage its Proton versions.",
        "",
        "For details, usage, and more, see the README.md file, or visit https://github.com/Mpaxlamitsounas/umu-commander.",
        sep="\n",
    )



import os
import sys
from collections.abc import Callable
from json import JSONDecodeError

from umu_commander import configuration as config
from umu_commander import database as db
from umu_commander import tracking, umu_config
from umu_commander.classes import ExitCode
from umu_commander.configuration import CONFIG_DIR, CONFIG_NAME
from umu_commander.util import print_help

# TODO: Add related projects shoutout
# https://github.com/Faugus/faugus-launcher
# https://github.com/SeongGino/Nero-umu
# https://github.com/korewaChino/umu-wrapper

# TODO: https://inquirerpy.readthedocs.io/en/latest/


def main() -> ExitCode:
    try:
        config.load()

    except (JSONDecodeError, KeyError):
        config_path: str = os.path.join(CONFIG_DIR, CONFIG_NAME)
        config_path_old: str = os.path.join(CONFIG_DIR, CONFIG_NAME + ".old")

        print(f"Config file at {config_path} could not be read.")

        if not os.path.exists(config_path_old):
            print(f"Config file renamed to {config_path_old}.")
            os.rename(config_path, config_path_old)

    except FileNotFoundError:
        config.dump()

    try:
        db.load()

    except JSONDecodeError:
        db_path: str = os.path.join(config.DB_DIR, config.DB_NAME)
        db_path_old: str = os.path.join(config.DB_DIR, config.DB_NAME + ".old")

        print(f"Tracking file at {db_path} could not be read.")

        if not os.path.exists(db_path_old):
            print(f"DB file renamed to {db_path_old}.")
            os.rename(db_path, db_path_old)

    except FileNotFoundError:
        pass

    dispatch: dict[str, Callable] = {
        "track": tracking.track,
        "untrack": tracking.untrack,
        "users": tracking.users,
        "delete": tracking.delete,
        "create": umu_config.create,
        "run": umu_config.run,
    }

    try:
        dispatch[sys.argv[1]]()

    except IndexError:
        print_help()
        return_val = ExitCode.SUCCESS

    except KeyError:
        print("Unrecognised verb.")
        print_help()
        return_val = ExitCode.INVALID_SELECTION

    except ValueError:
        return_val = ExitCode.INVALID_SELECTION

    else:
        return_val = ExitCode.SUCCESS

    finally:
        tracking.untrack_unlinked()
        db.dump()

    return return_val.value


if __name__ == "__main__":
    exit(main())



import json
import os
from collections import defaultdict

import umu_commander.configuration as config

_db: defaultdict[str, defaultdict[str, list[str]]] = defaultdict(
    lambda: defaultdict(list)
)


def load():
    global _db

    if not os.path.exists(config.DB_DIR):
        os.mkdir(config.DB_DIR)

    with open(
        os.path.join(os.path.join(config.DB_DIR, config.DB_NAME)), "rt"
    ) as db_file:
        _db.update(json.load(db_file))


def dump():
    if not os.path.exists(config.DB_DIR):
        os.mkdir(config.DB_DIR)

    with open(os.path.join(config.DB_DIR, config.DB_NAME), "wt") as db_file:
        # noinspection PyTypeChecker
        json.dump(_db, db_file, indent="\t")


def get(
    proton_dir: str = None, proton_ver: str = None
) -> dict[str, dict[str, list[str]]] | dict[str, list[str]] | list[str]:
    global _db

    if proton_dir is None and proton_ver is None:
        return _db

    if proton_ver is None:
        return _db[proton_dir]

    if proton_ver not in _db[proton_dir]:
        _db[proton_dir][proton_ver] = []

    return _db[proton_dir][proton_ver]


def _reset():
    global _db
    _db = defaultdict(lambda: defaultdict(list))



import importlib
import os
import tomllib
from pathlib import Path
from typing import Any

import tomli_w

from umu_commander.classes import DLLOverride

CONFIG_DIR: str = os.path.join(Path.home(), ".config")
CONFIG_NAME: str = "umu-commander.toml"


PROTON_PATHS: tuple[str, ...] = (
    os.path.join(Path.home(), ".local/share/Steam/compatibilitytools.d/"),
    os.path.join(Path.home(), ".local/share/umu/compatibilitytools"),
)
UMU_PROTON_PATH: str = os.path.join(
    Path.home(), ".local/share/Steam/compatibilitytools.d/"
)
DB_NAME: str = "tracking.json"
DB_DIR: str = os.path.join(Path.home(), ".local/share/umu/compatibilitytools")
UMU_CONFIG_NAME: str = "umu-config.toml"
DEFAULT_PREFIX_DIR: str = os.path.join(Path.home(), ".local/share/wineprefixes/")
DLL_OVERRIDES_OPTIONS: tuple[DLLOverride, ...] = (
    DLLOverride("winhttp for BepInEx", "winhttp.dll=n;"),
)


def load():
    with open(os.path.join(CONFIG_DIR, CONFIG_NAME), "rb") as conf_file:
        toml_conf = tomllib.load(conf_file)
        if "DLL_OVERRIDES_OPTIONS" in toml_conf:
            toml_conf["DLL_OVERRIDES_OPTIONS"] = tuple(
                [
                    DLLOverride(label, override_str)
                    for label, override_str in toml_conf[
                        "DLL_OVERRIDES_OPTIONS"
                    ].items()
                ]
            )

        module = importlib.import_module(__name__)
        for key, value in toml_conf.items():
            setattr(module, key, value)


def _get_attributes() -> dict[str, Any]:
    module = importlib.import_module(__name__)
    attributes: dict[str, Any] = {}
    for key in dir(module):
        value = getattr(module, key)
        if not key.startswith("_") and not callable(value) and key.upper() == key:
            attributes[key] = value

    return attributes


def dump():
    if not os.path.exists(CONFIG_DIR):
        os.mkdir(CONFIG_DIR)

    with open(os.path.join(CONFIG_DIR, CONFIG_NAME), "wb") as conf_file:
        toml_conf = _get_attributes()
        del toml_conf["CONFIG_DIR"]
        del toml_conf["CONFIG_NAME"]

        toml_conf["DLL_OVERRIDES_OPTIONS"] = dict(
            [(override.info, override.value) for override in DLL_OVERRIDES_OPTIONS]
        )

        tomli_w.dump(toml_conf, conf_file)



import os
import re
import subprocess

import umu_commander.configuration as config
import umu_commander.database as db
from umu_commander.classes import ProtonDir, ProtonVer


def _natural_sort_proton_ver_key(p: ProtonVer, _nsre=re.compile(r"(\d+)")):
    s: str = p.version_num
    return [int(text) if text.isdigit() else text for text in _nsre.split(s)]


def refresh_proton_versions():
    print("Updating umu Proton.")
    umu_update_process = subprocess.run(
        ["umu-run", '""'],
        env={"PROTONPATH": "UMU-Latest", "UMU_LOG": "debug"},
        capture_output=True,
        text=True,
    )

    for line in umu_update_process.stderr.split("\n"):
        if "PROTONPATH" in line and "/" in line:
            try:
                left: int = line.rfind("/") + 1
                print(f"Latest UMU-Proton: {line[left:len(line) - 1]}.")
            except ValueError:
                print("Could not fetch latest UMU-Proton.")

            break


def _sort_proton_versions(versions: list[ProtonVer]) -> list[ProtonVer]:
    return sorted(versions, key=_natural_sort_proton_ver_key, reverse=True)


def collect_proton_versions(
    sort: bool = False, user_count: bool = False
) -> list[ProtonDir]:
    def get_user_count(proton_dir: str, proton_ver) -> str:
        return (
            "(" + str(len(db.get(proton_dir, proton_ver))) + ")"
            if proton_ver in db.get(proton_dir)
            else "(-)"
        )

    proton_dirs: list[ProtonDir] = []
    for proton_dir in config.PROTON_PATHS:
        versions: list[ProtonVer] = [
            ProtonVer(
                proton_dir,
                version,
                get_user_count(proton_dir, version) if user_count else "",
            )
            for version in os.listdir(proton_dir)
            if os.path.isdir(os.path.join(proton_dir, version))
        ]

        if sort:
            versions = sorted(versions, key=_natural_sort_proton_ver_key, reverse=True)

        proton_dirs.append(
            ProtonDir(proton_dir, f"Proton versions in {proton_dir}:", versions)
        )

    return proton_dirs


def get_latest_umu_proton():
    umu_proton_versions: list[ProtonVer] = [
        ProtonVer(config.UMU_PROTON_PATH, version)
        for version in os.listdir(config.UMU_PROTON_PATH)
        if "UMU" in version
        and os.path.isdir(os.path.join(config.UMU_PROTON_PATH, version))
    ]
    umu_proton_versions = sorted(
        umu_proton_versions, key=_natural_sort_proton_ver_key, reverse=True
    )

    return umu_proton_versions[0].version_num
